import java.util.Random;  //<>//
import java.util.Arrays;

private final int SIM_COUNT = 1000;
private final int MAX_DEPTH = 6;
private final int POPULATION = 2000;//50;  // 1世代当たりの個体数
private final float SURVIVE_RATE = 0.8;   // 生き残り割合
private final float MUTATION_RATE = 0.8;//0.05; // 突然変異発生率
private Random rnd = new Random();
private Node[] cars = new Node[POPULATION];
private float[] score = new float[POPULATION];

float xval;


private enum NodeType {
  ADD("(+)"), SUB("(-)"), MUL("(*)"), DIV("(/)"), 
    ONE("[1]"), TWO("[2]"), SENSOR("[S]");
  private final String str;
  private NodeType(final String aStr) {
    str = aStr;
  }
  public String text() {
    return str;
  }
}

private class Node {
  public NodeType type;        // このノードの種別
  public Node child0, child1;  // 左右の枝 (自身が葉の場合は null のこと)

  public Node(NodeType aType, Node aChild0, Node aChild1) {
    rnd = new Random();
    type = aType;
    child0 = aChild0;
    child1 = aChild1;
  }

  public Node(int depth) {    // ランダムな式を生成する
    NodeType[] v = NodeType.values();
    if (depth == 0) {
      type = (new NodeType[] {NodeType.ADD, NodeType.SUB, 
        NodeType.MUL, NodeType.DIV})[rnd.nextInt(4)];
    } else if (depth < MAX_DEPTH) {
      type = v[rnd.nextInt(v.length)];
    } else {
      type = (new NodeType[] {NodeType.ONE, NodeType.TWO, 
        NodeType.SENSOR})[rnd.nextInt(3)];
    }
    if (type == NodeType.ADD || type == NodeType.SUB
      || type == NodeType.MUL || type == NodeType.DIV) {
      child0 = new Node(depth + 1);
      child1 = new Node(depth + 1);
    } else {
      child0 = null;
      child1 = null;
    }
  }

  // 突然変異 (node type をランダムに変化させる)
  public void mutation() {
    Node tgt = pickNodeRnd();
    if (tgt.child0 != null) {  // 葉ではない (2項演算子である)
      tgt.type = (new NodeType[] {NodeType.ADD, NodeType.SUB, NodeType.MUL, NodeType.DIV})[rnd.nextInt(4)];
    } else {                   // 葉である  (オペランドである)
      tgt.type = (new NodeType[] {NodeType.ONE, NodeType.TWO, NodeType.SENSOR})[rnd.nextInt(3)];
    }
  }

  public float evaluate () {
    float value = 0;
    switch (type) {
    case TWO:
      value = 2;
      break;
    case ONE:
      value = 1;
      break;
    case SENSOR:
      value = xval;
      break;
    case ADD:
      value = child0.evaluate() + child1.evaluate();
      break;
    case SUB:
      value = child0.evaluate() - child1.evaluate();
      break;
    case MUL:
      value = child0.evaluate() * child1.evaluate();
      break;
    case DIV:
      if (child1.evaluate() != 0) {
        value = child0.evaluate() / child1.evaluate();
      } else {
        // 0の除算は -1 倍扱いにすることでコード数節約兼エラー回避
        value = -child0.evaluate();
      }
      break;
    }
    return value;
  }

  // 自身の部分木(の根)をランダムに返す (交叉に使用)
  // (ランダムに掘り進んだ経路に対して一様分布で選ぶ。)
  public Node pickNodeRnd() {
    Node n = this;
    ArrayList <Node> nodes = new ArrayList <Node> ();
    while (n != null) {
      nodes.add(n);
      n = (rnd.nextInt(2) == 0) ? n.child0 : n.child1;
    }
    return nodes.get(rnd.nextInt(nodes.size() - 1) + 1);
  }

  // 自身の deep copy 生成 (突然変異のときは shallow copy だとまずい)
  @Override
    public Node clone() {
    if (child0 == null) return new Node(type, null, null);
    else return new Node(type, child0.clone(), child1.clone());
  }

  // debug よう
  public ArrayList<String> show() {
    ArrayList<String> str = new ArrayList<String>();
    str.add("-" + type.text());
    if (child0 != null) {
      for (String s : child0.show()) {
        str.add("  |" + s);
      }
      str.add("  |");
      for (String s : child1.show()) {
        str.add("   " + s);
      }
    }
    return str;
  }
  // 後置記法で文字列化 (保存よう)
  public String showRPN() {
    if (child0 != null) {
      return child0.showRPN() + " " + child1.showRPN() + " " + type.text();
    } else {
      return type.text();
    }
  }
  // 中置記法で文字列化
  public String showInfix() {
    String str = null;
    switch (type) {
    case TWO:
    case ONE:
    case SENSOR:
      str = (type.text()).substring(1, 2);
      break;
    case ADD:
    case SUB:
      str = String.format("(%s %s %s)", 
        child0.showInfix(), 
        (type.text()).substring(1, 2), 
        child1.showInfix());
      break;
    case DIV:
    case MUL:
      if (type == NodeType.DIV && child1.evaluate() == 0) { // 0の除算の例外
        str = String.format("%s * (-1)", child0.showInfix());
      } else {
        str = String.format("%s %s %s", 
          child0.showInfix(), 
          (type.text()).substring(1, 2), 
          child1.showInfix());
      }
      break;
    }
    return str;
  }
}


// 2つの親を交差させて2つの子([0][1]の配列)を作る
public static Node[] crossover(Node father, Node mother) {
  Node children[] = new Node[2];
  children[0] = father.clone();
  children[1] = mother.clone();
  Node n0 = children[0].pickNodeRnd();
  Node n1 = children[1].pickNodeRnd();
  // n0 と n1 を入れ替える
  NodeType tmpType = n0.type;
  Node     tmpChild0 = n0.child0;
  Node     tmpChild1 = n0.child1;
  n0.type = n1.type;
  n0.child0 = n1.child0;
  n0.child1 = n1.child1;
  n1.type = tmpType;
  n1.child0 = tmpChild0;
  n1.child1 = tmpChild1;
  return children;
}


void Init()
{
  gCount = 1;  // 世代カウンタ
  cCount = 0;  // (その世代中の) 個体カウンタ
  // 第一世代を初期化　
  for (int i = 0; i < POPULATION; i++) {
    cars[i] = new Node(0);
    score[i] = 0;
  }
}


void Evaluate()
{
  for (cCount =0; cCount< POPULATION; cCount++) {  // 1世代終了
    float y, val;
    score[cCount]=0;
    for (float x=-3; x<5; x+=1) {
      y = -3*x*x*x*x+4*x*x*x+12*x*x;
      xval=x;
      val = cars[cCount].evaluate();
      score[cCount] += (y-val)*(y-val);
    }
  }
}

void nextGeneration() {
  // スコア順に整列
  for (int i = 0; i < POPULATION - 1; i++) { // bubble sort
    for (int j = i + 1; j < POPULATION; j++) {
      //      if (score[i] < score[j]) {
      if (score[i] > score[j]) {
        float tmps = score[i];
        score[i] = score[j];
        score[j] = tmps;
        Node tmpc = cars[i];
        cars[i] = cars[j];
        cars[j] = tmpc;
      }
    }
  }
  // 最優秀コード/最低ノード発表
  println("*** C H A M P I O N   o f   T H I S  G E N . ***");
  writer_log.println(gCount+"\t"+score[0]);
  writer_log.println(cars[0].showInfix());     // 中置記法で保存
  for (String s : cars[0].show()) {
    writer_log.println(s);//階層構造を保存
  }
  writer.println(gCount+"\t"+score[0]+"\t\""+cars[0].showRPN()+"\"");//後置記法で文字列化 (保存よう)
  println(score[0]);
  // 上位を残し、残りはそれらの交配で充当
  int threshold = (int)(POPULATION * SURVIVE_RATE);
  for (int i = threshold; i < POPULATION - 1; i++) {
    // 交叉で 2 つ作られるが一つは捨てる
    cars[i] = crossover(cars[rnd.nextInt(threshold)], 
      cars[rnd.nextInt(threshold)])[0];
    if (Math.random() < MUTATION_RATE) {
      cars[i].mutation();
    }
  }
  cars[POPULATION - 1] = new Node(0);
  // スコアゼロクリア
  for (int i = 0; i < POPULATION; i++) {
    score[i] = 0;
  }
  gCount++;
}